/* xml++.cc
 * libxml++ and this file are copyright (C) 2000 by Ari Johnson, and
 * are covered by the GNU Lesser General Public License, which should be
 * included with libxml++ as the file COPYING.
 */

#include "libxml++/parsers/domparser.h"
#include "libxml++/dtd.h"
#include "libxml++/nodes/element.h"
#include "libxml++/nodes/textnode.h"
#include "libxml++/nodes/commentnode.h"
#include "libxml++/keepblanks.h"
#include "libxml++/exceptions/internal_error.h"
#include <libxml/parserInternals.h>//For xmlCreateFileParserCtxt().

#include <sstream>
#include <iostream>

namespace xmlpp
{

DomParser::DomParser()
: _doc(0)
{
  //Start with an empty document:
  _doc = new Document();
}

DomParser::DomParser(const std::string& filename, bool validate)
: _doc(0)
{
  set_validate(validate);
  parse_file(filename);
}

DomParser::~DomParser()
{ 
  release_underlying();
}

void DomParser::parse_file(const std::string& filename)
{
  release_underlying(); //Free any existing document.

  KeepBlanks k(KeepBlanks::Default);

  //The following is based on the implementation of xmlParseFile(), in xmlSAXParseFileWithData():
  _context = xmlCreateFileParserCtxt(filename.c_str());

  if(!_context)
  {
    throw internal_error("Couldn't create parsing context");
  }

  if(_context->directory == 0)
  {
    char* directory = xmlParserGetDirectory(filename.c_str());
    _context->directory = (char*) xmlStrdup((xmlChar*) directory);
  }

  parse_context();
}

void DomParser::parse_memory(const std::string& contents)
{
  release_underlying(); //Free any existing document.

  KeepBlanks k(KeepBlanks::Default);

  //The following is based on the implementation of xmlParseFile(), in xmlSAXParseFileWithData():
  _context = xmlCreateMemoryParserCtxt(contents.c_str(), contents.size());

  if(!_context)
  {
    throw internal_error("Couldn't create parsing context");
  }

  parse_context();
}

void DomParser::parse_context()
{
  KeepBlanks k(KeepBlanks::Default);

  //The following is based on the implementation of xmlParseFile(), in xmlSAXParseFileWithData():
  //and the implementation of xmlParseMemory(), in xmlSaxParseMemoryWithData().
  initialize_context();

  xmlParseDocument(_context);

  check_for_exception();

  if(!_context->wellFormed)
  {
    release_underlying(); //Free _doc;
    throw parse_error("Document not well-formed.");
  }

  if(_context->errNo != 0)
  {
    release_underlying();
    std::ostringstream o;
    o << "libxml error " << _context->errNo;
    throw parse_error(o.str());
  }

  _doc = new Document(_context->myDoc);

  //Free the parse context, but keep the document alive so people can navigate the DOM tree:
  //TODO: Why not keep the context alive too?
  Parser::release_underlying();

  check_for_exception();
}


void DomParser::parse_stream(std::istream& in)
{
  release_underlying(); //Free any existing document.

  KeepBlanks k(KeepBlanks::Default);

  _context = xmlCreatePushParserCtxt(
      0, // setting thoses two parameters to 0 force the parser
      0, // to create a document while parsing.
      0,
      0,
      ""); // here should come the filename. I don't know if it is a problem to let it empty

  if(!_context)
  {
    throw internal_error("Couldn't create parsing context");
  }

  initialize_context();

  std::string line;
  while(std::getline(in, line))
  {
    // since getline does not get the line separator, we have to add it since the parser cares
    // about layout in certain cases.
    line += '\n';

    xmlParseChunk(_context, line.c_str(), line.length(), 0);
  }

  xmlParseChunk(_context, 0, 0, 1);

  check_for_exception();

  if(!_context->wellFormed)
  {
    release_underlying(); //Free _doc;
    throw parse_error("Document not well-formed.");
  }

  if(_context->errNo != 0)
  {
    std::ostringstream o;
    o << "libxml error " << _context->errNo;
    release_underlying();
    throw parse_error(o.str());
  }

  _doc = new Document(_context->myDoc);

  //Free the parse context, but keep the document alive so people can navigate the DOM tree:
  //TODO: Why not keep the context alive too?
  Parser::release_underlying();

  check_for_exception();
}

void DomParser::release_underlying()
{
  if(_doc)
  {
    delete _doc;
    _doc = 0;
  }

  Parser::release_underlying();
}

DomParser::operator bool() const
{
  return _doc != 0;
}

Document* DomParser::get_document()
{
  return _doc;
}

const Document* DomParser::get_document() const
{
  return _doc;
}

} // namespace xmlpp


